﻿using System;
using System.IO.Ports;

namespace CPUCoolerTester
{
    public class Wrapper
    {
        public SerialPort SerialPort;

        public Wrapper(string portName)
        {
            try
            {
                SerialPort = new SerialPort(portName, 38400, Parity.None, 8, StopBits.One);
                SerialPort.WriteTimeout = 1000;
                SerialPort.ReadTimeout = 1000;
                SerialPort.Open();
                SerialPort.ReadExisting();
                SerialPort.WriteLine("V");
                if (!SerialPort.ReadLine().StartsWith("CPUCoolerTester 1.")) throw new InvalidResponseException();
            }
            catch
            {
                try
                {
                    SerialPort.Close();
                }
                catch { }
                throw;
            }
        }

        public class InvalidResponseException : Exception { }

        public void Reset()
        {
            SerialPort.WriteLine("R");
            if (SerialPort.ReadLine() != "A") { throw new InvalidResponseException(); };
        }

        public void TestMode()
        {
            SerialPort.WriteLine("T");
            if (SerialPort.ReadLine() != "A") { throw new InvalidResponseException(); };
        }

        public void SetConfig(UInt16 targetPower, UInt16 fanPWM)
        {
            SerialPort.WriteLine(String.Format("C{0:X4}{1:X4}", targetPower, fanPWM));
            if (SerialPort.ReadLine() != "A") { throw new InvalidResponseException(); };
        }

        public enum State : UInt16
        {
            OK = 0x0000,

            ALARM_WATCHDOG_TIMEOUT = 0x0001,
            ALARM_V12_OUT_OF_RANGE = 0x0002,
            ALARM_OVERHEAT = 0x0003,
            ALARM_ADC_OVERRANGE = 0x0004,
            ALARM_I2C_ERROR_ADC1 = 0x0005,
            ALARM_I2C_ERROR_ADC2 = 0x0006,
            ALARM_I2C_ERROR_ADC3 = 0x0007,
            ALARM_DAC_TEST_MODE = 0x0008,

            ALARM_HEATER1_VS_OUT_OF_RANGE = 0x0009,
            ALARM_HEATER1_OVER_CURRENT = 0x0010,
            ALARM_HEATER1_UNCONTROLLABLE = 0x0011,

            ALARM_HEATER2_VS_OUT_OF_RANGE = 0x0109,
            ALARM_HEATER2_OVER_CURRENT = 0x0110,
            ALARM_HEATER2_UNCONTROLLABLE = 0x0111,

            ALARM_HEATER3_VS_OUT_OF_RANGE = 0x0209,
            ALARM_HEATER3_OVER_CURRENT = 0x0210,
            ALARM_HEATER3_UNCONTROLLABLE = 0x0211,
        }

        public class HeaterStatus
        {
            public UInt16 DAC;
            public double VS;
            public double Current;
            public double Power;
        }

        public class Status
        {
            public UInt16 Timestamp;
            public State State;
            public UInt16 FanRPM_500ms;
            public UInt16 FanRPM_30s;
            public double Power;
            public double Temp;
            public double V12;
            public HeaterStatus[] Heater;
        }

        public static Double GetDouble(string s, ref int i)
        {
            var result = Convert.ToInt16(s.Substring(i, 4), 16);
            i += 4;

            if (result == Int16.MinValue)
            {
                return Double.NaN;
            }
            else
            {
                return result;
            }
        }

        public static UInt16 GetUInt16(string s, ref int i)
        {
            var result = Convert.ToUInt16(s.Substring(i, 4), 16);
            i += 4;
            return result;
        }

        public static double GetPower(string s, ref int i)
        {
            return 0.1 * GetDouble(s, ref i);
        }

        public static double GetTemperature(string s, ref int i)
        {
            return 0.01 * GetDouble(s, ref i);
        }

        public static double GetVoltage(string s, ref int i)
        {
            return 0.001 * GetDouble(s, ref i);
        }

        public static double GetCurrent(string s, ref int i)
        {
            return 0.001 * GetDouble(s, ref i);
        }

        public const int HEATER_COUNT = 3;

        public Status GetStatus()
        {
            SerialPort.WriteLine("S");
            var s = SerialPort.ReadLine();

            var data = new Status();
            try
            {
                var i = 0;
                data.Timestamp = GetUInt16(s, ref i);
                data.State = (State)GetUInt16(s, ref i);
                data.FanRPM_500ms = GetUInt16(s, ref i);
                data.FanRPM_30s = GetUInt16(s, ref i);
                data.Power = GetPower(s, ref i);
                data.Temp = GetTemperature(s, ref i);
                data.V12 = GetVoltage(s, ref i);

                data.Heater = new HeaterStatus[3];
                for (int j = 0; j < HEATER_COUNT; j++)
                {
                    var heaterStatus = new HeaterStatus();

                    heaterStatus.DAC = GetUInt16(s, ref i);
                    heaterStatus.VS = GetVoltage(s, ref i);
                    heaterStatus.Current = GetCurrent(s, ref i);
                    heaterStatus.Power = GetPower(s, ref i);

                    data.Heater[j] = heaterStatus;
                }
            }
            catch { throw new InvalidResponseException(); }

            return data;
        }
    }
}
